<?php


class StorageRackspace extends StorageContainer implements StorageContainerInterface {
  private $conn;

  private function log_exception(Exception $e, $dsm = TRUE) {
    watchdog('storage_rackspace', 'php-cloudfiles exception: @msg', array('@msg' => $e->getMessage()), WATCHDOG_ERROR);
    
    if ($dsm && user_access('administer storage')) {
      drupal_set_message(t('php-cloudfiles exception: @msg', array('@msg' => $e->getMessage())), 'error');
    }
  }

  private function init() {
    if (module_exists('libraries') && function_exists('libraries_load')) {
      $library = libraries_load('php-cloudfiles');
    }

    if (!isset($library) || $library['loaded'] == FALSE) {
      watchdog('storage_rackspace', 'php-cloudfiles not found', array(), WATCHDOG_ERROR);
      
      if (user_access('administer storage')) {
        drupal_set_message(t('php-cloudfiles not found. Check the <a href="!path">status report</a>.', 
          array('!path' => url('admin/reports/status'))), 'error');
      }
      
      throw new StorageException();
    }

    $auth = new CF_Authentication($this->settings['username'], $this->settings['api_key']);
    $auth->ssl_use_cabundle();

    try {
      $auth->authenticate();
    }
    catch (AuthenticationException $e) {
      $this->log_exception($e, FALSE);
      
      if (user_access('administer storage')) {
        drupal_set_message(t('Invalid username or access key.'), 'error');
      }
        
      throw new StorageException();
    }
    catch (Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }

    $this->conn = new CF_Connection($auth);
    $this->conn->ssl_use_cabundle();
  }

  private function cf_container() {
    $cf_containers = &drupal_static(__FUNCTION__, array());
    
    if (isset($cf_containers[$this->container_id])) {
      return $cf_containers[$this->container_id];
    }

    $this->init();
    
    try {
      $cf_container = $this->conn->get_container($this->settings['cf_container']);
    }
    catch (Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }

    $cf_containers[$this->container_id] = $cf_container;

    return $cf_container;
  }

  public function serviceSettingsForm($edit = FALSE) {
    // Add defaults.
    $this->settings += array(
      'username' => '',
      'cf_container' => '',
    );

    $items['credentials_info'] = array(
      '#markup' => t('Get your credentials !link.', array('!link' => l(t('here'), 
        'http://www.rackspace.com/cloud/cloud_hosting_products/files/', 
        array('attributes' => array('target' => '_blank')))))
    );

    $items['username'] = array(
      '#type' => 'textfield',
      '#title' => t('Username'),
      '#required' => TRUE,
      '#validated' => TRUE,
      '#default_value' => $this->settings['username'],
    );
    
    $items['api_key'] = array(
      '#type' => 'password',
      '#title' => t('API key'),
      '#required' => TRUE,
      '#validated' => TRUE,
    );

    if (!$edit) {
      $items['cf_container'] = array(
        '#type' => 'textfield',
        '#title' => t('Cloud Files container'),
        '#required' => TRUE,
        '#validated' => TRUE,
        '#default_value' => $this->settings['cf_container'],
      );
    }
    
    return $items;
  }
  
  public function serviceSettingsValidate($edit = FALSE) {
    if (!$this->settings['username']) {
      form_set_error('settings][username', t('Username field is required.'));
    }

    if (!$this->settings['api_key']) {
      form_set_error('settings][api_key', t('API key field is required.'));
    }
    
    if (!$this->settings['cf_container']) {
      form_set_error('settings][cf_container', t('Cloud Files container name field is required.'));
    }
    
    if (form_get_errors()) {
      return;
    }
    
    try {
      $this->init();
    }
    catch (StorageException $e) {
      form_set_error('settings');
    }
  }
  
  public function serviceCreate() {
    $this->init();
    
    try {
      $cf_container = $this->conn->get_container($this->settings['cf_container']);
    }
    catch (NoSuchContainerException $e) {}
    catch (Exception $e) {
      $this->log_exception($e);
      form_set_error('settings');
      throw new StorageException();
    }
    
    if (isset($cf_container)) {
      form_set_error('settings][cf_container', t('Cloud Files container %name already exists.', 
        array('%name' => $this->settings['cf_container'])));
      throw new StorageException();
    }
    
    try {
      $cf_container = $this->conn->create_container($this->settings['cf_container']);
    
      if ($this->access_control) {
        $cf_container->make_private();    // strangely, this has to be done explicity, or it might be public
      }
      else {
        $cf_container->make_public(60 * 60 * 24 * 3);    // 3 days is the maximum ttl that can be set in the api
        $this->settings['base_url'] = $cf_container->cdn_uri;
      }
    }
    catch (Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }
  }
  
  public function serviceInfo() {
    $info = array(
      t('Username') => $this->settings['username'],
      t('Container') => $this->settings['cf_container'],
    );

    return $info;
  }
  
  public function serviceDestroy() {
    $this->init();
    
    try {
      $this->conn->delete_container($this->settings['cf_container']);
    }
    catch (Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }
  }
  
  public function serviceInstanceCreate(Storage $storage, $uri) {
    $reference = $this->uniqueFilename($storage->filename);

    $cf_container = $this->cf_container();

    try {
      $object = $cf_container->create_object($reference);
      $object->content_type = $storage->mimetype;
      $object->set_etag(base64_encode($storage->md5));
      $object->load_from_filename($uri, FALSE);

      $object->metadata = array(
        'storage_id' => $storage->storage_id,
        'file_id' => $storage->file_id,
        'filename' => $storage->filename,
      );

      $object->sync_metadata();
    }
    catch(Exception $e) {
      log_exception($e);

      try {
        $cf_container->delete_object($reference);
      }
      catch(Exception $e) {}

      throw new StorageException();
    }

    return $reference;
  }
  
  public function serviceInstanceExists($reference) {
    $cf_container = $this->cf_container();

    try {
      $object = $cf_container->get_object($reference);
    }
    catch(Exception $e) {
      return FALSE;
    }

    return TRUE;
  }

  public function serviceInstanceDestroy($reference) {
    $cf_container = $this->cf_container();
      
    try {
      $cf_container->delete_object($reference);
    }
    catch(Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }
  }
  
  public function serviceInstanceGetUri($reference) {
    $cf_container = $this->cf_container();
    $temp_uri = storage_temp_uri();
    
    try {
      $object = $cf_container->get_object($reference);
      $object->save_to_filename($temp_uri);
    }
    catch(Exception $e) {
      $this->log_exception($e);
      throw new StorageException();
    }

    return $temp_uri;
  }
}

